from setuptools import setup

setup(
    name="arrdem.datalog",
    # Package metadata
    version="0.0.7",
    license="MIT",
    description="A Datalog engine",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    author="Reid 'arrdem' McKenzie",
    author_email="me@arrdem.com",
    url="https://git.arrdem.com/arrdem/datalog-py",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Topic :: Database",
        "Topic :: Database :: Database Engines/Servers",
        "Topic :: Database :: Front-Ends",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
    ],

    # Package setup
    package_dir={"": "src"},
    packages=[
        "datalog",
    ],
    scripts=["bin/datalog"],
    install_requires=[
        "yaspin==0.14.3",
    ],
)
